'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = parseCommandLineOptions;

var _commander = require('commander');

var _commander2 = _interopRequireDefault(_commander);

var _package = require('../package.json');

var _package2 = _interopRequireDefault(_package);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _builtinTransforms = require('./builtinTransforms');

var _builtinTransforms2 = _interopRequireDefault(_builtinTransforms);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var transformsDocs = '\n  Safe transforms:\n\n    + arrow .......... callback to arrow function\n    + for-of ......... for loop to for-of loop\n    + for-each ....... for loop to Array.forEach()\n    + arg-rest ....... use of arguments to function(...args)\n    + arg-spread ..... use of apply() to spread operator\n    + obj-method ..... function values in objects to methods\n    + obj-shorthand .. {foo: foo} to {foo}\n    + no-strict ...... remove "use strict" directives\n    + commonjs ....... CommonJS module loading to import/export\n    + exponent ....... Math.pow() to ** operator (ES7)\n    + multi-var ...... single var x,y; declaration to var x; var y; (refactor)\n\n  Unsafe transforms:\n\n    + let ............ var to let/const\n    + class .......... prototype assignments to class declaration\n    + template ....... string concatenation to template string\n    + default-param .. use of || to default parameters\n    + includes ....... indexOf() != -1 to includes() (ES7)\n';

_commander2.default.usage('-t <transform> <file>');
_commander2.default.description(_package2.default.description + '\n' + transformsDocs);
_commander2.default.version(_package2.default.version);
_commander2.default.option('-o, --out-file <file>', 'write output to a file');
_commander2.default.option('--replace <dir>', 'in-place transform all *.js files in a directory\n                         <dir> can also be a single file or a glob pattern');
_commander2.default.option('-t, --transform <a,b,c>', 'one or more transformations to perform', function (v) {
  return v.split(',');
});

/**
 * Parses and validates command line options from argv.
 *
 * - On success returns object with options.
 * - On failure throws exceptions with error message to be shown to user.
 *
 * @param {String[]} argv Raw command line arguments
 * @return {Object} options object
 */
function parseCommandLineOptions(argv) {
  _commander2.default.parse(argv);
  return {
    inFile: getInputFile(),
    outFile: _commander2.default.outFile,
    replace: getReplace(),
    transforms: getTransforms()
  };
}

function getInputFile() {
  if (_commander2.default.args.length > 1) {
    throw 'Only one input file allowed, but ' + _commander2.default.args.length + ' given instead.';
  }
  if (_commander2.default.args[0] && !_fs2.default.existsSync(_commander2.default.args[0])) {
    throw 'File ' + _commander2.default.args[0] + ' does not exist.';
  }
  return _commander2.default.args[0];
}

function getReplace() {
  if (!_commander2.default.replace) {
    return undefined;
  }
  if (_commander2.default.outFile) {
    throw 'The --replace and --out-file options cannot be used together.';
  }
  if (_commander2.default.args[0]) {
    throw 'The --replace and plain input file options cannot be used together.\n' + 'Did you forget to quote the --replace parameter?';
  }
  if (_fs2.default.existsSync(_commander2.default.replace) && _fs2.default.statSync(_commander2.default.replace).isDirectory()) {
    return _path2.default.join(_commander2.default.replace, '/**/*.js');
  }
  return _commander2.default.replace;
}

function getTransforms() {
  if (!_commander2.default.transform || _commander2.default.transform.length === 0) {
    throw 'No transforms specifed :(\n\nUse --transform option to pick one of the following:\n' + transformsDocs;
  }

  // Ensure only valid transform names are used
  validateTransforms(_commander2.default.transform);

  return _commander2.default.transform;
}

function validateTransforms(transformNames) {
  transformNames.forEach(function (name) {
    if (!_builtinTransforms2.default.get(name)) {
      throw 'Unknown transform "' + name + '".';
    }
  });
}
module.exports = exports['default'];