#!/usr/bin/env node
/**
 * Command line parse utility.
 * Read from STDIN, write to STDOUT.
 */

'use strict';

require('../core-upgrade.js');

var ParserEnv = require('../lib/config/MWParserEnvironment.js').MWParserEnvironment;
var ParsoidConfig = require('../lib/config/ParsoidConfig.js').ParsoidConfig;
var TemplateRequest = require('../lib/mw/ApiRequest.js').TemplateRequest;
var Util = require('../lib/utils/Util.js').Util;
var DU = require('../lib/utils/DOMUtils.js').DOMUtils;
var Promise = require('../lib/utils/promise.js');
var fs = require('fs');
var path = require('path');
var yargs = require('yargs');
var yaml = require('js-yaml');

var initTime;

process.on('SIGUSR2', function() {
	var heapdump = require('heapdump');
	console.error('SIGUSR2 received! Writing snapshot.');
	process.chdir('/tmp');
	heapdump.writeSnapshot();
});

var standardOpts = Util.addStandardOptions({
	'wt2html': {
		description: 'Wikitext -> HTML',
		'boolean': true,
		'default': false,
	},
	'html2wt': {
		description: 'HTML -> Wikitext',
		'boolean': true,
		'default': false,
	},
	'wt2wt': {
		description: 'Wikitext -> HTML -> Wikitext',
		'boolean': true,
		'default': false,
	},
	'html2html': {
		description: 'HTML -> Wikitext -> HTML',
		'boolean': true,
		'default': false,
	},
	'selser': {
		description: 'Use the selective serializer to go from HTML to Wikitext.',
		'boolean': true,
		'default': false,
	},
	'normalize': {
		description: 'Normalize the output as parserTests would do. Use --normalize for PHP tests, and --normalize=parsoid for parsoid-only tests',
		'default': false,
	},
	'config': {
		description: "Path to a config.yaml file.  Use --config w/ no argument to default to the server's config.yaml",
		'default': false,
	},
	'prefix': {
		description: 'Which wiki prefix to use; e.g. "enwiki" for English wikipedia, "eswiki" for Spanish, "mediawikiwiki" for mediawiki.org',
		'boolean': false,
		'default': null,
	},
	'domain': {
		description: 'Which wiki to use; e.g. "en.wikipedia.org" for English wikipedia, "es.wikipedia.org" for Spanish, "mediawiki.org" for mediawiki.org',
		'boolean': false,
		'default': null,
	},
	'page': {
		description: 'The page name, returned for {{PAGENAME}}. If no input is given (ie. empty/stdin closed), it downloads and parses the page. This should be the actual title of the article (that is, not including any URL-encoding that might be necessary in wikitext).',
		'boolean': false,
		'default': ParserEnv.prototype.defaultPageName,
	},
	'contentmodel': {
		description: 'The content model of the input.  Defaults to "wikitext" but extensions may support others (for example, "json").',
		'boolean': false,
		'default': null,
	},
	'oldid': {
		description: 'Oldid of the given page.',
		'boolean': false,
		'default': null,
	},
	'oldtext': {
		description: 'The old page text for a selective-serialization (see --selser)',
		'boolean': false,
		'default': null,
	},
	'oldtextfile': {
		description: 'File containing the old page text for a selective-serialization (see --selser)',
		'boolean': false,
		'default': null,
	},
	'oldhtmlfile': {
		description: 'File containing the old HTML for a selective-serialization (see --selser)',
		'boolean': false,
		'default': null,
	},
	'domdiff': {
		description: 'File containing the diff-marked HTML for used with selective-serialization (see --selser)',
		'boolean': false,
		'default': null,
	},
	'inputfile': {
		description: 'File containing input as an alternative to stdin',
		'boolean': false,
		'default': false,
	},
	'contentVersion': {
		description: 'The acceptable content version.',
		'boolean': false,
		'default': ParserEnv.prototype.contentVersion,
	},
	'pagebundle': {
		description: 'Output pagebundle JSON',
		'boolean': true,
		'default': false,
	},
	'pbin': {
		description: 'Input pagebundle JSON',
		'boolean': false,
		'default': '',
	},
	'pbinfile': {
		description: 'Input pagebundle JSON file',
		'boolean': false,
		'default': '',
	},
	'pboutfile': {
		description: 'Output pagebundle JSON to file',
		'boolean': false,
		'default': false,
	},
	'lint': {
		description: 'Parse with linter enabled',
		'boolean': true,
		'default': false,
	},
	'scrubWikitext': {
		description: 'Apply wikitext scrubbing while serializing.',
		'boolean': true,
		'default': false,
	},
	'nativeGallery': {
		description: 'Omit extsrc from gallery.',
		'boolean': true,
		'default': false,
	},
	'loadWMF': {
		description: 'Use WMF mediawiki API config',
		'boolean': true,
		'default': true,
	},
	'offline': {
		description: 'Shortcut to turn off various network fetches during parse.',
		'boolean': true,
		'default': false,
	},
	'record': {
		description: 'Record http requests for later replay',
		'boolean': true,
		'default': false,
	},
	'replay': {
		description: 'Replay recorded http requests for later replay',
		'boolean': true,
		'default': false,
	},
	'useBatchAPI': {
		description: 'Turn on/off the API batching system',
		// Since I picked a null default (to let the default config setting be the default),
		// I cannot make this a boolean option.
		'boolean': false,
		'default': null,
	},
});
exports.defaultOptions = yargs.options(standardOpts).parse([]);

var startsAtWikitext;
var startsAtHTML = function(argv, env, input, pb) {
	var doc = DU.parseHTML(input);
	pb = pb || DU.extractPageBundle(doc);
	if (argv.selser) {
		pb = pb || DU.extractPageBundle(env.page.dom.ownerDocument);
		if (pb) {
			DU.applyPageBundle(env.page.dom.ownerDocument, pb);
		}
	}
	if (pb) {
		DU.applyPageBundle(doc, pb);
	}
	var handler = env.getContentHandler(argv.contentmodel);
	return handler.fromHTML(env, doc.body, argv.selser)
	.then(function(out) {
		if (argv.html2wt || argv.wt2wt) {
			return { trailingNL: true, out: out, env: env };
		} else {
			return startsAtWikitext(argv, env, out);
		}
	});
};

startsAtWikitext = function(argv, env, input) {
	env.setPageSrcInfo(input);
	var handler = env.getContentHandler(argv.contentmodel);
	return handler.toHTML(env)
	.tap(function(doc) {
		if (env.conf.parsoid.useBatchAPI) {
			return DU.addRedLinks(env, doc);
		}
	})
	.then(function(doc) {
		if (argv.lint) {
			env.log("end/parse");
		}
		if (argv.wt2html || argv.html2html) {
			var out;
			if (argv.pboutfile) {
				var pb = DU.extractPageBundle(doc);
				fs.writeFileSync(argv.pboutfile, JSON.stringify(pb), 'utf8');
			}
			if (argv.normalize) {
				out = DU.normalizeOut(doc.body, (argv.normalize === 'parsoid'));
			} else if (argv.document) {
				// used in Parsoid JS API, return document
				out = doc;
			} else {
				out = DU.toXML(doc);
			}
			return { trailingNL: true, out: out, env: env };
		} else {
			return startsAtHTML(argv, env, DU.toXML(doc));
		}
	});
};

var parse = exports.parse = function(input, argv, parsoidConfig, prefix, domain) {
	var env;
	var s1, s2;
	s1 = Date.now();
	return ParserEnv.getParserEnv(parsoidConfig, {
		prefix: prefix,
		domain: domain,
		pageName: argv.page,
	}).then(function(_env) {
		env = _env;
		env.startTime = s1;
		if (initTime) {
			// When used in lib/index.js as an exported module,
			// this will be undefined. And, since we will be
			// may be using this in a repl setup, init time is
			// not as relevant.
			env.bumpTimeUse("Init", initTime);
		}

		// fetch templates from enwiki by default.
		if (argv.wgScriptPath) {
			env.conf.wiki.wgScriptPath = argv.wgScriptPath;
		}

		// Enable wikitext scrubbing
		env.scrubWikitext = argv.scrubWikitext;

		// Enable html editing galleries
		env.nativeGallery = argv.nativeGallery;

		// Sets ids on nodes and stores data-* attributes in a JSON blob
		env.pageBundle = argv.pagebundle || argv.pboutfile;

		// The content version to output
		if (argv.contentVersion) {
			env.setContentVersion(argv.contentVersion);
		}

		if (!argv.wt2html) {
			if (argv.oldtextfile) {
				argv.oldtext = fs.readFileSync(argv.oldtextfile, 'utf8');
			}
			if (argv.oldhtmlfile) {
				env.page.dom = DU.parseHTML(
					fs.readFileSync(argv.oldhtmlfile, 'utf8')
				).body;
			}
			if (argv.domdiff) {
				// FIXME: need to load diff markers from attributes
				env.page.domdiff = {
					isEmpty: false,
					dom: DU.ppToDOM(fs.readFileSync(argv.domdiff, 'utf8')),
				};
				throw new Error('this is broken');
			}
			env.setPageSrcInfo(argv.oldtext || null);
		}

		if (argv.selser && argv.oldtext === null) {
			throw new Error('Please provide original wikitext ' +
				'(--oldtext or --oldtextfile). Selser requires that.');
		}

		if (typeof input === 'string') {
			return input;
		}

		s2 = Date.now();
		env.bumpTimeUse("Setup Environment", s2 - s1);

		if (argv.inputfile) {
			// read input from the file, then process
			var fileContents = fs.readFileSync(argv.inputfile, 'utf8');
			return fileContents;
		}

		// Send a message to stderr if there is no input for a while, since the
		// convention that --page must be used with </dev/null is confusing.
		var stdinTimer = setTimeout(function() {
			console.error('Waiting for stdin...');
		}, 1000);

		return new Promise(function(resolve) {
			// collect input
			var inputChunks = [];
			var stdin = process.stdin;
			stdin.resume();
			stdin.setEncoding('utf8');
			stdin.on('data', function(chunk) {
				inputChunks.push(chunk);
			});
			stdin.on('end', function() {
				resolve(inputChunks);
			});
		}).then(function(inputChunks) {
			clearTimeout(stdinTimer);
			// parse page if no input
			if (inputChunks.length > 0) {
				return inputChunks.join('');
			} else if (argv.html2wt || argv.html2html) {
				env.log("fatal", "Pages start at wikitext.");
			}
			var target = env.normalizeAndResolvePageTitle();
			return TemplateRequest
				.setPageSrcInfo(env, target, argv.oldid)
				.then(function() {
					// Preserve fetched contentmodel.
					argv.contentmodel = argv.contentmodel ||
						env.page.meta.revision.contentmodel;
					return env.page.src;
				});
		});
	}).then(function(str) {
		str = str.replace(/\r/g, '');
		if (argv.html2wt || argv.html2html) {
			var pb;
			if (argv.pbin.length > 0) {
				pb = JSON.parse(argv.pbin);
			} else if (argv.pbinfile) {
				pb = JSON.parse(fs.readFileSync(argv.pbinfile, 'utf8'));
			}
			return startsAtHTML(argv, env, str, pb);
		} else {
			env.bumpTimeUse("Pre-parse (source fetch)", Date.now() - s2);
			return startsAtWikitext(argv, env, str);
		}
	});
};

if (require.main === module) {
	(function() {
		var start = Date.now();
		var defaultModeStr = "Default conversion mode : --wt2html";

		var opts = yargs.usage(
			'Usage: echo wikitext | $0 [options]\n\n' + defaultModeStr,
			standardOpts
		).strict();

		var argv = opts.parse(process.argv);

		if (Util.booleanOption(argv.help)) {
			opts.showHelp();
			return;
		}

		// Because selser builds on html2wt serialization,
		// the html2wt flag should be automatically set when selser is set.
		if (argv.selser) {
			argv.html2wt = true;
		}

		// Default conversion mode
		if (!argv.html2wt && !argv.wt2wt && !argv.html2html) {
			argv.wt2html = true;
		}

		var prefix = argv.prefix || null;
		var domain = argv.domain || null;

		if (argv.apiURL) {
			prefix = 'customwiki';
			domain = null;
		} else if (!(prefix || domain)) {
			domain = 'en.wikipedia.org';
		}

		var parsoidOptions = {
			loadWMF: argv.loadWMF,
		};

		if (Util.booleanOption(argv.config)) {
			var p = (typeof (argv.config) === 'string') ?
				path.resolve('.', argv.config) :
				path.resolve(__dirname, '../config.yaml');
			// Assuming Parsoid is the first service in the list
			parsoidOptions = yaml.load(fs.readFileSync(p, 'utf8')).services[0].conf;
		}

		Util.setTemplatingAndProcessingFlags(parsoidOptions, argv);
		Util.setDebuggingFlags(parsoidOptions, argv);

		// Offline shortcut
		if (argv.offline) {
			parsoidOptions.fetchConfig = false;
			parsoidOptions.fetchTemplates = false;
			parsoidOptions.fetchImageInfo = false;
			parsoidOptions.usePHPPreProcessor = false;
			parsoidOptions.expandExtensions = false;
		}

		if (parsoidOptions.localsettings) {
			parsoidOptions.localsettings = path.resolve(__dirname, parsoidOptions.localsettings);
		}

		var pc = new ParsoidConfig(null, parsoidOptions);
		pc.defaultWiki = prefix ? prefix : pc.reverseMwApiMap.get(domain);

		var nock, dir, nocksFile;
		if (argv.record || argv.replay) {
			prefix = prefix || 'enwiki';
			dir = path.resolve(__dirname, '../nocks/');
			if (!fs.existsSync(dir)) {
				fs.mkdirSync(dir);
			}
			dir = dir + '/' + prefix;
			if (!fs.existsSync(dir)) {
				fs.mkdirSync(dir);
			}
			nocksFile = dir + '/' + argv.page + '.js';
		}

		if (argv.record) {
			nock = require('nock');
			nock.recorder.rec({ dont_print: true });
		} else if (argv.replay) {
			try {
				require(nocksFile);
			} catch (e) {
				console.error('Exception ' + e + ' requiring ' + nocksFile);
				console.error('Cannot replay!');
				return -1;
			}
		}

		initTime = Date.now() - start;

		return parse(null, argv, pc, prefix, domain)
		.then(function(res) {
			var stdout = process.stdout;
			stdout.write(res.out);
			if (res.trailingNL && stdout.isTTY) {
				stdout.write('\n');
			}
			if (argv.record) {
				var nockCalls = nock.recorder.play();
				var stream = fs.createWriteStream(nocksFile);
				stream.once('open', function() {
					stream.write("var nock = require('nock');");
					for (var i = 0; i < nockCalls.length; i++) {
						stream.write(nockCalls[i]);
					}
					stream.end();
				});
			}
		}).done();
	}());
}
